/*
 * Copyright (c) 2016 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#include <stdlib.h>
#include <stdatomic.h>
#include <math.h>
#include <inttypes.h>

#include <tracker.h>

#include <timely.lv2/timely.h>

typedef struct _plughandle_t plughandle_t;

struct _plughandle_t {
	LV2_URID_Map *map;
	LV2_Atom_Forge forge;
	LV2_Atom_Forge_Ref ref;

	LV2_Log_Log *log;
	LV2_Log_Logger logger;

	const LV2_Atom_Sequence *control;
	LV2_Atom_Sequence *notify;

	PROPS_T(props, MAX_NPROPS);
	plugstate_t state;
	plugstate_t stash;
	uint32_t mode_size;
	uint32_t graph_size;

	LV2_URID midi_MidiEvent;
	LV2_URID tracker_position;

	timely_t timely;
	bool rolling;
	int64_t offset;
	int64_t last;
	LV2_Atom_Event *ev [MAX_TRACKS];
};

static void
_all_notes_off(plughandle_t *handle, int64_t frames)
{
	for(unsigned c = 0; c < 16; c++)
	{
		const uint8_t msg [3] = {
			LV2_MIDI_MSG_CONTROLLER | c,
			LV2_MIDI_CTL_ALL_NOTES_OFF,
			0x0
		};

		if(handle->ref)
			handle->ref = lv2_atom_forge_frame_time(&handle->forge, frames);
		if(handle->ref)
			handle->ref = lv2_atom_forge_atom(&handle->forge, sizeof(msg), handle->midi_MidiEvent);
		if(handle->ref)
			handle->ref = lv2_atom_forge_write(&handle->forge, msg, sizeof(msg));
	}
}

static inline LV2_Atom_Event *
_reposition_track(plughandle_t *handle, const LV2_Atom_Sequence *seq)
{
	LV2_ATOM_SEQUENCE_FOREACH(seq, ev)
	{
		const int64_t beat_frames = ev->time.beats * TIMELY_FRAMES_PER_BEAT(&handle->timely);

		if(beat_frames >= handle->offset)
			return ev; // reposition here
	}

	return NULL;
}

static inline void
_reposition(plughandle_t *handle)
{
	timely_t *timely = &handle->timely;

	handle->rolling = TIMELY_SPEED(timely) > 0.f ? true : false;

	// current beat time minus tracker offset
	const double beats = (double)TIMELY_BAR(timely) * TIMELY_BEATS_PER_BAR(timely)
		+ TIMELY_BAR_BEAT_RAW(timely) - (handle->state.offset - 1);

	handle->state.position = fmod(beats, handle->state.beats);
	if(handle->state.repeats > 0) // indefinite repeats
	{
		const double rep = beats / handle->state.beats; // repeats already done so far

		if(rep >= handle->state.repeats) // advance position after repeats have overflown
			handle->state.position += floor(rep) * handle->state.beats;
	}

	handle->offset = handle->state.position * TIMELY_FRAMES_PER_BEAT(timely) - 1;

	unsigned track = 0;
	LV2_ATOM_TUPLE_BODY_FOREACH(handle->state.graph, handle->graph_size, item)
	{
		const LV2_Atom_Sequence *seq = (const LV2_Atom_Sequence *)item;

		handle->ev[track] = _reposition_track(handle, seq);

		if(++track >= MAX_TRACKS)
			break;
	}
}

static void
_intercept(void *data, int64_t frames, props_impl_t *impl)
{
	plughandle_t *handle = data;

	_reposition(handle);
}

static void
_intercept_mode(void *data, int64_t frames, props_impl_t *impl)
{
	plughandle_t *handle = data;

	handle->mode_size = impl->value.size;
}

static void
_intercept_graph(void *data, int64_t frames, props_impl_t *impl)
{
	plughandle_t *handle = data;

	handle->graph_size = impl->value.size;
	_reposition(handle);

	_all_notes_off(handle, frames);
}

static const props_def_t defs [MAX_NPROPS] = {
	{
		.property = TRACKER__offset,
		.offset = offsetof(plugstate_t, offset),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept,
	},
	{
		.property = TRACKER__beats,
		.offset = offsetof(plugstate_t, beats),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept,
	},
	{
		.property = TRACKER__fractions,
		.offset = offsetof(plugstate_t, fractions),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept,
	},
	{
		.property = TRACKER__repeats,
		.offset = offsetof(plugstate_t, repeats),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept,
	},
	{
		.property = TRACKER__position,
		.offset = offsetof(plugstate_t, position),
		.type = LV2_ATOM__Double,
		.access = LV2_PATCH__readable
	},
	{
		.property = TRACKER__mode,
		.offset = offsetof(plugstate_t, mode),
		.type = LV2_ATOM__Tuple,
		.max_size = MODE_SIZE,
		.event_cb = _intercept_mode,
	},
	{
		.property = TRACKER__graph,
		.offset = offsetof(plugstate_t, graph),
		.type = LV2_ATOM__Tuple,
		.max_size = GRAPH_SIZE,
		.event_cb = _intercept_graph,
	}
};

static inline void
_play(plughandle_t *handle, int64_t to)
{
	const int64_t rel = handle->offset - to; // beginning of current period

	while(true)
	{
		int64_t next_frames = INT64_MAX;
		int next_track = -1;

		int track = 0;
		LV2_ATOM_TUPLE_BODY_FOREACH(handle->state.graph, handle->graph_size, item)
		{
			const LV2_Atom_Sequence *seq = (const LV2_Atom_Sequence *)item;

			LV2_Atom_Event *ev = handle->ev[track];
			if(ev && !lv2_atom_sequence_is_end(&seq->body, seq->atom.size, ev))
			{
				const int64_t beat_frames = ev->time.beats * TIMELY_FRAMES_PER_BEAT(&handle->timely);

				if(beat_frames >= handle->offset)
				{
					goto cont; // event not part of this period
				}

#if 0
				lv2_log_trace(&handle->logger, "next event: %lf %li %li\n",
					ev->time.beats, beat_frames, handle->offset);
#endif

				const int64_t frames = beat_frames - rel;

				if(frames < next_frames)
				{
					next_frames = frames;
					next_track = track;
				}
			}

cont:
			if(++track >= MAX_TRACKS)
				break;
		}

		if(next_track >= 0) // more scheduled events to process
		{
			LV2_Atom_Event *ev = handle->ev[next_track];

			// check for time jump! skip out-of-order event, as it probably has already been forged...
			if(next_frames >= handle->last) //TODO can this be solved more elegantly?
			{
				// append events
				if(ev->body.type == handle->forge.Tuple) // multiple events per slot
				{
					const LV2_Atom_Tuple *tup = (const LV2_Atom_Tuple *)&ev->body;

					LV2_ATOM_TUPLE_FOREACH(tup, atom)
					{
						if(handle->ref)
							handle->ref = lv2_atom_forge_frame_time(&handle->forge, next_frames);
						if(handle->ref)
							handle->ref = lv2_atom_forge_write(&handle->forge, atom, lv2_atom_total_size(atom));
					}
				}
				else // !Tuple
				{
					const LV2_Atom *atom = &ev->body;

					if(handle->ref)
						handle->ref = lv2_atom_forge_frame_time(&handle->forge, next_frames);
					if(handle->ref)
						handle->ref = lv2_atom_forge_write(&handle->forge, atom, lv2_atom_total_size(atom));
				}

				handle->last = next_frames; // advance frame time head
			}

			// advance event pointer on given track
			handle->ev[next_track] = lv2_atom_sequence_next(ev);
		}
		else // no more scheduled events to preocess
		{
			break;
		}
	}
}

static void
_timely_cb(timely_t *timely, int64_t frames, LV2_URID type, void *data)
{
	plughandle_t *handle = data;

	_reposition(handle);
}

static LV2_Handle
instantiate(const LV2_Descriptor* descriptor, double rate,
	const char *bundle_path, const LV2_Feature *const *features)
{
	plughandle_t *handle = calloc(1, sizeof(plughandle_t));
	if(!handle)
		return NULL;

	for(unsigned i=0; features[i]; i++)
	{
		if(!strcmp(features[i]->URI, LV2_URID__map))
			handle->map = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_LOG__log))
			handle->log = features[i]->data;
	}

	if(!handle->map)
	{
		fprintf(stderr,
			"%s: Host does not support urid:map\n", descriptor->URI);
		free(handle);
		return NULL;
	}

	if(handle->log)
		lv2_log_logger_init(&handle->logger, handle->map, handle->log);

	lv2_atom_forge_init(&handle->forge, handle->map);

	handle->tracker_position = handle->map->map(handle->map->handle, TRACKER__position);
	handle->midi_MidiEvent = handle->map->map(handle->map->handle, LV2_MIDI__MidiEvent);

	if(!props_init(&handle->props, descriptor->URI,
		defs, MAX_NPROPS, &handle->state, &handle->stash,
		handle->map, handle))
	{
		fprintf(stderr, "props_init failed\n");
		free(handle);
		return NULL;
	}

	const timely_mask_t mask = TIMELY_MASK_BAR_BEAT
		//| TIMELY_MASK_BAR
		| TIMELY_MASK_BEAT_UNIT
		| TIMELY_MASK_BEATS_PER_BAR
		| TIMELY_MASK_BEATS_PER_MINUTE
		| TIMELY_MASK_FRAMES_PER_SECOND
		| TIMELY_MASK_SPEED
		| TIMELY_MASK_BAR_BEAT_WHOLE;
	timely_init(&handle->timely, handle->map, rate, mask, _timely_cb, handle);

	// init state
	handle->state.offset = 1;
	handle->state.beats = 4;
	handle->state.fractions = 8;
	handle->state.repeats = 0;
	memcpy(&handle->stash, &handle->state, sizeof(plugstate_t));

	return handle;
}

static void
connect_port(LV2_Handle instance, uint32_t port, void *data)
{
	plughandle_t *handle = instance;

	switch(port)
	{
		case 0:
			handle->control = data;
			break;
		case 1:
			handle->notify = data;
			break;
	}
}

static void
run(LV2_Handle instance, uint32_t nsamples)
{
	plughandle_t *handle = instance;

	handle->last = 0; // reset frame time head

	const uint32_t capacity = handle->notify->atom.size;
	LV2_Atom_Forge_Frame frame;
	lv2_atom_forge_set_buffer(&handle->forge, (uint8_t *)handle->notify, capacity);
	handle->ref = lv2_atom_forge_sequence_head(&handle->forge, &frame, 0);

	props_idle(&handle->props, &handle->forge, 0, &handle->ref);

	int64_t last_t = 0;
	LV2_ATOM_SEQUENCE_FOREACH(handle->control, ev)
	{
		const LV2_Atom *atom= &ev->body;
		const LV2_Atom_Object *obj = (const LV2_Atom_Object *)&ev->body;

		if(handle->rolling)
			handle->offset += ev->time.frames - last_t;

		if(  !timely_advance(&handle->timely, obj, last_t, ev->time.frames)
			&& !props_advance(&handle->props, &handle->forge, ev->time.frames, obj, &handle->ref) )
		{
			// pass unknown events through, e.g. for rack-based hosts
			const uint32_t sz = lv2_atom_total_size(atom);
			if(handle->ref)
				handle->ref = lv2_atom_forge_frame_time(&handle->forge, ev->time.frames);
			if(handle->ref)
				handle->ref = lv2_atom_forge_write(&handle->forge, atom, sz);
		}

		if(handle->rolling)
			_play(handle, ev->time.frames);

		last_t = ev->time.frames;
	}

	{
		if(handle->rolling)
			handle->offset += nsamples - last_t;

		timely_advance(&handle->timely, NULL, last_t, nsamples);

		if(handle->rolling)
			_play(handle, nsamples);
	}

	{
		const double position = (double)handle->offset / TIMELY_FRAMES_PER_BEAT(&handle->timely);
		const double old_beat = handle->state.position * handle->state.fractions;
		const double new_beat = position * handle->state.fractions;

		const bool sync = (new_beat == old_beat) || (floor(new_beat) != floor(old_beat));

		handle->state.position = position;

		if(sync)
			props_set(&handle->props, &handle->forge, nsamples - 1, handle->tracker_position, &handle->ref);
	}

	if(handle->ref)
		lv2_atom_forge_pop(&handle->forge, &frame);
	else
		lv2_atom_sequence_clear(handle->notify);
}

static void
cleanup(LV2_Handle instance)
{
	plughandle_t *handle = instance;

	free(handle);
}

static LV2_State_Status
_state_save(LV2_Handle instance, LV2_State_Store_Function store,
	LV2_State_Handle state, uint32_t flags, const LV2_Feature *const *features)
{
	plughandle_t *handle = instance;

	return props_save(&handle->props, store, state, flags, features);
}

static LV2_State_Status
_state_restore(LV2_Handle instance, LV2_State_Retrieve_Function retrieve,
	LV2_State_Handle state, uint32_t flags, const LV2_Feature *const *features)
{
	plughandle_t *handle = instance;

	return props_restore(&handle->props, retrieve, state, flags, features);
}

static const LV2_State_Interface state_iface = {
	.save = _state_save,
	.restore = _state_restore
};

static const void*
extension_data(const char* uri)
{
	if(!strcmp(uri, LV2_STATE__interface))
		return &state_iface;

	return NULL;
}

static const LV2_Descriptor tracker_tracker = {
	.URI            = TRACKER_PREFIX"tracker",
	.instantiate    = instantiate,
	.connect_port   = connect_port,
	.activate       = NULL,
	.run            = run,
	.deactivate     = NULL,
	.cleanup        = cleanup,
	.extension_data = extension_data
};

static const LV2_Descriptor tracker_sequencer = {
	.URI            = TRACKER_PREFIX"sequencer",
	.instantiate    = instantiate,
	.connect_port   = connect_port,
	.activate       = NULL,
	.run            = run,
	.deactivate     = NULL,
	.cleanup        = cleanup,
	.extension_data = extension_data
};

LV2_SYMBOL_EXPORT const LV2_Descriptor*
lv2_descriptor(uint32_t index)
{
	switch(index)
	{
		case 0:
			return &tracker_tracker;
		case 1:
			return &tracker_sequencer;
		default:
			return NULL;
	}
}
