/*
 * Copyright (c) 2016 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#include <math.h>
#include <inttypes.h>

#include <tracker.h>

#include <d2tk/frontend_pugl.h>

#define MAX_COLS 128 //FIXME
#define MAX_ROWS 1024 //FIXME

typedef struct _row_t row_t;
typedef struct _plughandle_t plughandle_t;

struct _row_t {
	bool note [MAX_COLS];
};

struct _plughandle_t {
	LV2_URID_Map *map;
	LV2_URID_Unmap *unmap;
	LV2_Atom_Forge forge;

	LV2_URID atom_eventTransfer;
	LV2_URID atom_beatTime;
	LV2_URID midi_MidiEvent;
	LV2_URID tracker_offset;
	LV2_URID tracker_beats;
	LV2_URID tracker_fractions;
	LV2_URID tracker_repeats;
	LV2_URID tracker_position;
	LV2_URID tracker_mode;
	LV2_URID tracker_graph;

	LV2_Log_Log *log;
	LV2_Log_Logger logger;

	d2tk_pugl_config_t config;
	d2tk_frontend_t *dpugl;

	LV2UI_Controller *controller;
	LV2UI_Write_Function writer;

	PROPS_T(props, MAX_NPROPS);
	plugstate_t state;
	plugstate_t stash;
	uint32_t mode_size;
	uint32_t graph_size;

	row_t matrix [MAX_ROWS];

	unsigned position;

	atom_ser_t ser;

	float sample_rate;

	bool is_down;
	unsigned J;
	unsigned I;

	bool drag;
	d2tk_pos_t from_pos;
	d2tk_pos_t to_pos;
};

static void
_set_property(plughandle_t *handle, LV2_URID property)
{
	atom_ser_t *ser = &handle->ser;
	ser->offset = 0;
	lv2_atom_forge_set_sink(&handle->forge, _sink, _deref, ser);

	LV2_Atom_Forge_Ref ref = 1;
	props_set(&handle->props, &handle->forge, 0, property, &ref);

	const LV2_Atom_Event *ev = ser->ev;
	const LV2_Atom *atom = &ev->body;
	handle->writer(handle->controller, 0, lv2_atom_total_size(atom),
		handle->atom_eventTransfer, atom);
}

static void
_save_state(plughandle_t *handle)
{
	LV2_Atom_Forge *forge = &handle->forge;
	LV2_Atom_Forge_Frame tup_frame;
	LV2_Atom_Forge_Frame seq_frame;

	atom_ser_t *ser = &handle->ser;

	// serialize graph
	ser->offset = 0;
	lv2_atom_forge_set_sink(forge, _sink, _deref, ser);
	lv2_atom_forge_tuple(forge, &tup_frame);
	lv2_atom_forge_sequence_head(forge, &seq_frame, handle->atom_beatTime);
	{
		for(int j = 0; j < MAX_ROWS; j++)
		{
			// first handle note on
			for(int i = 0; i < MAX_COLS; i++)
			{
				const bool prev = (j == 0)
					? false
					: handle->matrix[j-1].note[i];

				if(handle->matrix[j].note[i] && !prev)
				{
					const uint8_t msg [3] = {
						LV2_MIDI_MSG_NOTE_ON | 0x0,
						i,
						0x7f
					};
					double beat_time = (j + 0.0) / handle->state.fractions;

					lv2_atom_forge_beat_time(forge, beat_time);
					lv2_atom_forge_atom(forge, sizeof(msg), handle->midi_MidiEvent);
					lv2_atom_forge_write(forge, msg, sizeof(msg));
					lv2_log_note(&handle->logger, "adding NoteOn: %lf\n", beat_time);
				}
			}

			// then handle note off
			for(int i = 0; i < MAX_COLS; i++)
			{
				const bool next = (j == MAX_ROWS - 1)
					? false
					: handle->matrix[j+1].note[i];

				if(handle->matrix[j].note[i] && !next)
				{
					const uint8_t msg [3] = {
						LV2_MIDI_MSG_NOTE_OFF | 0x0,
						i,
						0x0
					};
					double beat_time = (j + 1 - 1.0/handle->sample_rate) / handle->state.fractions;

					lv2_atom_forge_beat_time(forge, beat_time);
					lv2_atom_forge_atom(forge, sizeof(msg), handle->midi_MidiEvent);
					lv2_atom_forge_write(forge, msg, sizeof(msg));
					lv2_log_note(&handle->logger, "adding NoteOff: %lf\n", beat_time);
				}
			}
		}
	}
	lv2_atom_forge_pop(forge, &seq_frame);
	lv2_atom_forge_pop(forge, &tup_frame);

	const LV2_Atom *atom = ser->atom;
	if(atom->size <= GRAPH_SIZE)
	{
		//FIXME
		props_impl_t *impl = _props_impl_get(&handle->props, handle->tracker_graph);
		if(impl)
		{
			impl->value.size = atom->size;
			memcpy(handle->state.graph, LV2_ATOM_BODY_CONST(atom), atom->size);
		}
	}

	// send all patch:Set messages
	_set_property(handle, handle->tracker_offset);
	_set_property(handle, handle->tracker_beats);
	_set_property(handle, handle->tracker_fractions);
	_set_property(handle, handle->tracker_repeats);
	_set_property(handle, handle->tracker_mode);
	_set_property(handle, handle->tracker_graph);
}

static void
_restore_state(plughandle_t *handle)
{
	atom_ser_t *ser = &handle->ser;
	ser->offset = 0;
	lv2_atom_forge_set_sink(&handle->forge, _sink, _deref, ser);

	LV2_Atom_Forge_Frame frame;
	lv2_atom_forge_object(&handle->forge, &frame, 0, handle->props.urid.patch_get);
	lv2_atom_forge_pop(&handle->forge, &frame);

	const LV2_Atom *atom = ser->atom;
	handle->writer(handle->controller, 0, lv2_atom_total_size(atom),
		handle->atom_eventTransfer, atom);
}

static void
_load_state(plughandle_t *handle, uint32_t size, const void *body)
{
	// clear matrix
	memset(handle->matrix, 0x0, sizeof(handle->matrix));

	LV2_ATOM_TUPLE_BODY_FOREACH(body, size, item)
	{
		const LV2_Atom_Sequence *seq = (const LV2_Atom_Sequence *)item;

		LV2_ATOM_SEQUENCE_FOREACH(seq, ev)
		{
			const LV2_Atom *atom = &ev->body;

			const int j = round(ev->time.beats * handle->state.fractions);

			if( (atom->size == 3) && (atom->type == handle->midi_MidiEvent) )
			{
				const uint8_t *msg = LV2_ATOM_BODY_CONST(atom);

				const uint8_t cmd = msg[0] & 0xf0;
				//const uint8_t cha = msg[0] & 0x0f;

				switch(cmd)
				{
					case LV2_MIDI_MSG_NOTE_ON:
					{
						const uint8_t note = msg[1];
						const int i = note;

						handle->matrix[j].note[i] = true;
					} break;
					case LV2_MIDI_MSG_NOTE_OFF:
					{
						const uint8_t note = msg[1];
						const int i = note;

						for(int k = j-1; k >= 0; k--)
						{
							if(handle->matrix[k].note[i]) // found NoteOn //FIXME
							{
								break;
							}

							handle->matrix[k].note[i] = true; //FIXME
						}
					} break;
				}
			}
		}
	}
}

static void
_intercept_position(void *data, int64_t frames, props_impl_t *impl)
{
	plughandle_t *handle = data;

	const unsigned position = handle->state.position * handle->state.fractions;

	if(position != handle->position)
	{
		d2tk_frontend_redisplay(handle->dpugl);
		handle->position = position;
	}
}

static void
_intercept_graph(void *data, int64_t frames, props_impl_t *impl)
{
	plughandle_t *handle = data;

	handle->graph_size = impl->value.size;
	_load_state(handle, handle->graph_size, handle->state.graph);
}

static const props_def_t defs [MAX_NPROPS] = {
	{
		.property = TRACKER__offset,
		.offset = offsetof(plugstate_t, offset),
		.type = LV2_ATOM__Int
	},
	{
		.property = TRACKER__beats,
		.offset = offsetof(plugstate_t, beats),
		.type = LV2_ATOM__Int,
		//FIXME .event_cb = _intercept
	},
	{
		.property = TRACKER__fractions,
		.offset = offsetof(plugstate_t, fractions),
		.type = LV2_ATOM__Int,
		//FIXME .event_cb = _intercept
	},
	{
		.property = TRACKER__repeats,
		.offset = offsetof(plugstate_t, repeats),
		.type = LV2_ATOM__Int
	},
	{
		.property = TRACKER__position,
		.offset = offsetof(plugstate_t, position),
		.type = LV2_ATOM__Double,
		.event_cb = _intercept_position
	},
	{
		.property = TRACKER__mode,
		.offset = offsetof(plugstate_t, mode),
		.type = LV2_ATOM__Tuple,
		.max_size = MODE_SIZE,
		//FIXME .event_cb = _intercept_mode
	},
	{
		.property = TRACKER__graph,
		.offset = offsetof(plugstate_t, graph),
		.type = LV2_ATOM__Tuple,
		.max_size = GRAPH_SIZE,
		.event_cb = _intercept_graph
	}
};

static const char *note_lbl [12] = {
	"C", "#",
	"D", "#",
	"E",
	"F", "#",
	"G", "#",
	"A", "#",
	"B"
};

static const char *octave_lbl [11] = {
	"-1", "+0", "+1", "+2", "+3", "+4", "+5", "+6", "+7", "+8", "+9"
};

static inline void
_render_seq(plughandle_t *handle, const d2tk_rect_t *rect)
{
	const int N = MAX_COLS;
	const int M = MAX_ROWS;
	const int R = handle->state.beats * handle->state.fractions;
	const int NO = rect->w / 16;
	const int MO = NO * rect->h / rect->w;
	d2tk_frontend_t *dpugl = handle->dpugl;
	d2tk_base_t *base = d2tk_frontend_get_base(dpugl);
	bool state_changed = false;

	const uint32_t max [2] = { N, R };
	const uint32_t num [2] = { NO, MO };
	D2TK_BASE_SCROLLBAR(base, rect, D2TK_ID, D2TK_FLAG_SCROLL, max, num, scrollbar)
	{
		const d2tk_rect_t *sub = d2tk_scrollbar_get_rect(scrollbar);
		const int offset_x = ceilf(d2tk_scrollbar_get_offset_x(scrollbar));
		const int offset_y = ceilf(d2tk_scrollbar_get_offset_y(scrollbar));

		d2tk_style_t style = *d2tk_base_get_default_style();
		style.border_width = 1;
		style.padding = 0;
		style.rounding = 0;
		d2tk_base_set_style(base, &style);

		D2TK_BASE_TABLE(sub, NO, MO, D2TK_FLAG_TABLE_REL, tab)
		{
			const int x = d2tk_table_get_index_x(tab);
			const int y = d2tk_table_get_index_y(tab);
			const d2tk_rect_t *bnd = d2tk_table_get_rect(tab);

			if( (x <= 1) && (y <= 1) )
			{
				continue;
			}

			const int i = x - 2 + offset_x;
			const int j = y - 2 + offset_y;

			const d2tk_id_t id = D2TK_ID_IDX(j*N + i);

			if( (j > M ) || (j >= R) )
			{
				break;
			}

			if(y == 0)
			{
				if( (i % 12) == 0)
				{
					const char *lbl = octave_lbl[i / 12];
					d2tk_base_label(base, -1, lbl, 0.7f, bnd, D2TK_ALIGN_CENTERED);
				}

				continue;
			}
			else if(y == 1)
			{
				const char *lbl = note_lbl[i % 12];
				d2tk_base_label(base, -1, lbl, 1.f, bnd, D2TK_ALIGN_CENTERED);

				continue;
			}

			if(x == 0)
			{
				if( (j % handle->state.fractions) == 0)
				{
					char lbl [4];
					const size_t lbl_len = snprintf(lbl, sizeof(lbl), "%03i", (j / handle->state.fractions + handle->state.offset) % 1000);
					d2tk_base_label(base, lbl_len, lbl, 0.7f, bnd, D2TK_ALIGN_RIGHT | D2TK_ALIGN_MIDDLE);
				}

				continue;
			}
			else if(x == 1)
			{
				char lbl [3];
				const size_t lbl_len = snprintf(lbl, sizeof(lbl), "%i", j % handle->state.fractions + 1);
				d2tk_base_label(base, lbl_len, lbl, 1.f, bnd, D2TK_ALIGN_CENTERED);

				continue;
			}

			const bool I = (i / 12) % 2;
			const bool J = (j / handle->state.fractions) % 2;

			const uint32_t fill_color_active = style.fill_color[D2TK_TRIPLE_ACTIVE];
			const uint32_t fill_color_active_hot = style.fill_color[D2TK_TRIPLE_ACTIVE_HOT];

			if(j == round(handle->state.position * handle->state.fractions) )
			{
				style.fill_color[D2TK_TRIPLE_NONE] = 0x5f5f5fff;
				style.fill_color[D2TK_TRIPLE_ACTIVE] = 0x7f7f00ff;
				style.fill_color[D2TK_TRIPLE_ACTIVE_HOT] = 0xffff00ff;
			}
			else if( (I && !J) || (!I && J) )
			{
				style.fill_color[D2TK_TRIPLE_NONE] = 0x4f4f4fff;
			}
			else
			{
				style.fill_color[D2TK_TRIPLE_NONE] = 0x3f3f3fff;
			}

			bool *val = &handle->matrix[j].note[i];
			bool clone = *val;
			const uint32_t col_active = style.fill_color[D2TK_TRIPLE_ACTIVE];

			if(handle->drag)
			{
				d2tk_coord_t y0 = handle->from_pos.y <= handle->to_pos.y ? handle->from_pos.y : handle->to_pos.y;
				d2tk_coord_t y1 = handle->from_pos.y >  handle->to_pos.y ? handle->from_pos.y : handle->to_pos.y;
				d2tk_coord_t x0 = handle->from_pos.x <= handle->to_pos.x ? handle->from_pos.x : handle->to_pos.x;
				d2tk_coord_t x1 = handle->from_pos.x >  handle->to_pos.x ? handle->from_pos.x : handle->to_pos.x;

				y0 -= offset_y - 2;
				y1 -= offset_y - 2;
				x0 -= offset_x - 2;
				x1 -= offset_x - 2;

				for(d2tk_coord_t Y = y0; Y <= y1; Y++)
				{
					for(d2tk_coord_t X = x0; X <= x1; X++)
					{
						if( (y == handle->from_pos.y) && (x == handle->from_pos.x) )
						{
							continue;
						}

						if( (y != Y) || (x != X) )
						{
							continue;
						}

						clone = true;

						style.fill_color[D2TK_TRIPLE_ACTIVE] = 0x9f00cfff;
						break;
					}
				}
			}

			const d2tk_state_t state = d2tk_base_toggle(base, id, bnd, &clone);

			style.fill_color[D2TK_TRIPLE_ACTIVE] = col_active;

			if(d2tk_state_is_changed(state))
			{
				*val = clone;
			}

			if(d2tk_state_is_down(state))
			{
				handle->drag = true;

				handle->from_pos.x = handle->to_pos.x = i;
				handle->from_pos.y = handle->to_pos.y = j;

				fprintf(stdout, "toggle %016"PRIx64" DOWN\n", id);
			}

			if(handle->drag && d2tk_state_is_over(state))
			{
				handle->to_pos.x = i;
				handle->to_pos.y = j;
			}

			if(d2tk_state_is_up(state))
			{
				handle->drag = false;

				const d2tk_coord_t y0 = handle->from_pos.y <= handle->to_pos.y ? handle->from_pos.y : handle->to_pos.y;
				const d2tk_coord_t y1 = handle->from_pos.y >  handle->to_pos.y ? handle->from_pos.y : handle->to_pos.y;
				const d2tk_coord_t x0 = handle->from_pos.x <= handle->to_pos.x ? handle->from_pos.x : handle->to_pos.x;
				const d2tk_coord_t x1 = handle->from_pos.x >  handle->to_pos.x ? handle->from_pos.x : handle->to_pos.x;

				for(d2tk_coord_t Y = y0; Y <= y1; Y++)
				{
					for(d2tk_coord_t X = x0; X <= x1; X++)
					{
						handle->matrix[Y].note[X] = *val;
					}
				}

				fprintf(stdout, "toggle %016"PRIx64" UP\n", id);
			}

			style.fill_color[D2TK_TRIPLE_ACTIVE] = fill_color_active;
			style.fill_color[D2TK_TRIPLE_ACTIVE_HOT] = fill_color_active_hot;
		}

		d2tk_base_set_default_style(base);
	}

	if(state_changed)
	{
		//FIXME mark dirty state
	}
}

static inline void
_render_props(plughandle_t *handle, const d2tk_rect_t *rect)
{
	d2tk_frontend_t *dpugl = handle->dpugl;
	d2tk_base_t *base = d2tk_frontend_get_base(dpugl);

	D2TK_BASE_TABLE(rect, 8, 1, D2TK_FLAG_TABLE_REL, tab)
	{
		const int k = d2tk_table_get_index(tab);
		const d2tk_rect_t *bnd = d2tk_table_get_rect(tab);
		const d2tk_id_t id = D2TK_ID_IDX(k);

		switch(k)
		{
			case 0:
			{
				d2tk_base_label(base, -1, "Offset", 0.5f, bnd,
					D2TK_ALIGN_RIGHT | D2TK_ALIGN_MIDDLE);
			} break;
			case 1:
			{
				if(d2tk_base_prop_int32_is_changed(base, id, bnd,
					1, &handle->state.offset, 1000))
				{
					lv2_log_note(&handle->logger, "offset: %"PRIi32"\n", handle->state.offset); //FIXME
				}
			} break;

			case 2:
			{
				d2tk_base_label(base, -1, "Beats", 0.5f, bnd,
					D2TK_ALIGN_RIGHT | D2TK_ALIGN_MIDDLE);
			} break;
			case 3:
			{
				if(d2tk_base_prop_int32_is_changed(base, id, bnd,
					1, &handle->state.beats, 1000))
				{
					lv2_log_note(&handle->logger, "beats: %"PRIi32"\n", handle->state.beats); //FIXME
				}
			} break;

			case 4:
			{
				d2tk_base_label(base, -1, "Fractions", 0.5f, bnd,
					D2TK_ALIGN_RIGHT | D2TK_ALIGN_MIDDLE);
			} break;
			case 5:
			{
				if(d2tk_base_prop_int32_is_changed(base, id, bnd,
					1, &handle->state.fractions, 256))
				{
					lv2_log_note(&handle->logger, "fractions: %"PRIi32"\n", handle->state.fractions); //FIXME
				}
			} break;

			case 6:
			{
				d2tk_base_label(base, -1, "Repeats", 0.5f, bnd,
					D2TK_ALIGN_RIGHT | D2TK_ALIGN_MIDDLE);
			} break;
			case 7:
			{
				if(d2tk_base_prop_int32_is_changed(base, id, bnd,
					0, &handle->state.repeats, 1024))
				{
					lv2_log_note(&handle->logger, "repeats: %"PRIi32"\n", handle->state.repeats); //FIXME
				}
			} break;
		}
	}
}

static inline void
_render_menu(plughandle_t *handle, const d2tk_rect_t *rect)
{
	d2tk_frontend_t *dpugl = handle->dpugl;
	d2tk_base_t *base = d2tk_frontend_get_base(dpugl);

	D2TK_BASE_TABLE(rect, 3, 1, D2TK_FLAG_TABLE_REL, tab)
	{
		const int k = d2tk_table_get_index(tab);
		const d2tk_rect_t *bnd = d2tk_table_get_rect(tab);
		const d2tk_id_t id = D2TK_ID_IDX(k);

		switch(k)
		{
			case 0:
			{
				if(d2tk_base_button_label_is_changed(base, id, -1, "Restore",
					D2TK_ALIGN_CENTERED, bnd))
				{
					lv2_log_note(&handle->logger, "restoring\n"); //FIXME
					_restore_state(handle);
				}
			} break;
			case 1:
			{
				if(d2tk_base_button_label_is_changed(base, id, -1, "Commit",
					D2TK_ALIGN_CENTERED, bnd))
				{
					lv2_log_note(&handle->logger, "commiting\n"); //FIXME
					_save_state(handle);
				}
			} break;
			case 2:
			{
				d2tk_base_label(base, -1, "Tracker#Sequencer "TRACKER_VERSION, 0.6f,
					bnd, D2TK_ALIGN_RIGHT | D2TK_ALIGN_BOTTOM);
			} break;
		}
	}
}

static int
_expose(void *data, d2tk_coord_t w, d2tk_coord_t h)
{
	plughandle_t *handle = data;
	const d2tk_coord_t frac [3] = { 30, 0, 30};

	D2TK_BASE_LAYOUT(&D2TK_RECT(0, 0, w, h), 3, frac, D2TK_FLAG_LAYOUT_Y_ABS, vlay)
	{
		const d2tk_rect_t *vrect = d2tk_layout_get_rect(vlay);
		const unsigned k = d2tk_layout_get_index(vlay);

		switch(k)
		{
			case 0:
			{
				_render_props(handle, vrect);
			} break;
			case 1:
			{
				_render_seq(handle, vrect);
			} break;
			case 2:
			{
				_render_menu(handle, vrect);
			} break;
		}
	}

	return 0;
}

static LV2UI_Handle
instantiate(const LV2UI_Descriptor *descriptor, const char *plugin_uri,
	const char *bundle_path, LV2UI_Write_Function write_function,
	LV2UI_Controller controller, LV2UI_Widget *widget,
	const LV2_Feature *const *features)
{
	plughandle_t *handle = calloc(1, sizeof(plughandle_t));
	if(!handle)
		return NULL;

	void *parent = NULL;
	LV2UI_Resize *host_resize = NULL;
	LV2_Options_Option *opts = NULL;
	for(int i=0; features[i]; i++)
	{
		if(!strcmp(features[i]->URI, LV2_UI__parent))
			parent = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_UI__resize))
			host_resize = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_URID__map))
			handle->map = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_URID__unmap))
			handle->unmap = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_LOG__log))
			handle->log = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_OPTIONS__options))
			opts = features[i]->data;
	}

	if(!parent)
	{
		fprintf(stderr,
			"%s: Host does not support ui:parent\n", descriptor->URI);
		free(handle);
		return NULL;
	}
	if(!handle->map)
	{
		fprintf(stderr,
			"%s: Host does not support urid:map\n", descriptor->URI);
		free(handle);
		return NULL;
	}

	if(handle->log)
		lv2_log_logger_init(&handle->logger, handle->map, handle->log);

	lv2_atom_forge_init(&handle->forge, handle->map);

	if(!props_init(&handle->props, plugin_uri,
		defs, MAX_NPROPS, &handle->state, &handle->stash,
		handle->map, handle))
	{
		fprintf(stderr, "props_init failed\n");
		free(handle);
		return NULL;
	}

	handle->atom_eventTransfer = handle->map->map(handle->map->handle, LV2_ATOM__eventTransfer);
	handle->atom_beatTime = handle->map->map(handle->map->handle, LV2_ATOM__beatTime);
	handle->midi_MidiEvent = handle->map->map(handle->map->handle, LV2_MIDI__MidiEvent);
	handle->tracker_offset = handle->map->map(handle->map->handle, TRACKER__offset);
	handle->tracker_beats = handle->map->map(handle->map->handle, TRACKER__beats);
	handle->tracker_fractions = handle->map->map(handle->map->handle, TRACKER__fractions);
	handle->tracker_repeats = handle->map->map(handle->map->handle, TRACKER__repeats);
	handle->tracker_position = handle->map->map(handle->map->handle, TRACKER__position);
	handle->tracker_mode = handle->map->map(handle->map->handle, TRACKER__mode);
	handle->tracker_graph = handle->map->map(handle->map->handle, TRACKER__graph);

	const LV2_URID params_sample_rate = handle->map->map(handle->map->handle,
		LV2_PARAMETERS__sampleRate);

	handle->sample_rate = 48000.f; // fall-back

	for(LV2_Options_Option *opt = opts;
		opt && (opt->key != 0) && (opt->value != NULL);
		opt++)
	{
		if( (opt->key == params_sample_rate) && (opt->type == handle->forge.Float) )
			handle->sample_rate = *(float*)opt->value;
		//TODO handle more options
	}

	handle->controller = controller;
	handle->writer = write_function;

	const d2tk_coord_t w = 800;
	const d2tk_coord_t h = 450;

	d2tk_pugl_config_t *config = &handle->config;
	config->parent = (uintptr_t)parent;
	config->bundle_path = bundle_path;
	config->min_w = w/2;
	config->min_h = h/2;
	config->w = w;
	config->h = h;
	config->fixed_size = false;
	config->fixed_aspect = false;
	config->expose = _expose;
	config->data = handle;

	handle->dpugl = d2tk_pugl_new(config, (uintptr_t *)widget);
	if(!handle->dpugl)
	{
		free(handle);
		return NULL;
	}

	if(host_resize)
	{
		host_resize->ui_resize(host_resize->handle, w, h);
	}

	//FIXME
	handle->state.offset = 1;
	handle->state.repeats = 0;

	atom_ser_t *ser = &handle->ser;
	ser->size = 1024;
	ser->buf = malloc(ser->size);

	_restore_state(handle);

	return handle;
}

static void
cleanup(LV2UI_Handle instance)
{
	plughandle_t *handle = instance;

	d2tk_frontend_free(handle->dpugl);

	atom_ser_t *ser = &handle->ser;
	if(ser->buf)
		free(ser->buf);

	free(handle);
}

static void
port_event(LV2UI_Handle instance, uint32_t index, uint32_t size,
	uint32_t protocol, const void *buf)
{
	plughandle_t *handle = instance;

	if(protocol == handle->atom_eventTransfer)
	{
		const LV2_Atom_Object *obj = buf;

		atom_ser_t *ser = &handle->ser;
		ser->offset = 0;
		lv2_atom_forge_set_sink(&handle->forge, _sink, _deref, ser);

		LV2_Atom_Forge_Ref ref = 0;
		if(props_advance(&handle->props, &handle->forge, 0, obj, &ref))
		{
#if 0
			nk_pugl_post_redisplay(&handle->win);
#endif
		}
	}
}

static int
_idle(LV2UI_Handle instance)
{
	plughandle_t *handle = instance;

	return d2tk_frontend_step(handle->dpugl);
}

static const LV2UI_Idle_Interface idle_ext = {
	.idle = _idle
};

static int
_resize(LV2UI_Handle instance, int width, int height)
{
	plughandle_t *handle = instance;

	return d2tk_frontend_set_size(handle->dpugl, width, height);
}

static const LV2UI_Resize resize_ext = {
	.ui_resize = _resize
};

static const void *
extension_data(const char *uri)
{
	if(!strcmp(uri, LV2_UI__idleInterface))
	{
		return &idle_ext;
	}
	else if(!strcmp(uri, LV2_UI__resize))
	{
		return &resize_ext;
	}
		
	return NULL;
}

const LV2UI_Descriptor tracker_sequencer_ui = {
	.URI            = TRACKER_PREFIX"sequencer_ui",
	.instantiate    = instantiate,
	.cleanup        = cleanup,
	.port_event     = port_event,
	.extension_data = extension_data
};
