/*
 * Copyright (c) 2016 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <limits.h>
#include <inttypes.h>

#include <tracker.h>

#include <lua.h>
#include <lauxlib.h>
#include <lualib.h>

#include <api_forge.h>
#include <api_atom.h>

#define NK_PUGL_API
#include "nk_pugl/nk_pugl.h"

#define EDIT_MAX 32

typedef struct _message_t message_t;
typedef struct _track_t track_t;
typedef struct _plughandle_t plughandle_t;

struct _message_t {
	LV2_Atom atom;
	uint8_t body [0];
};

struct _track_t {
	int mode;
	message_t **messages;
	lua_State *L;
	int ref; // reference to registry of global Lua state
};

struct _plughandle_t {
	LV2_URID_Map *map;
	LV2_URID_Unmap *unmap;
	LV2_Atom_Forge forge;

	LV2_URID atom_eventTransfer;
	LV2_URID atom_beatTime;
	LV2_URID midi_MidiEvent;
	LV2_URID tracker_offset;
	LV2_URID tracker_beats;
	LV2_URID tracker_fractions;
	LV2_URID tracker_repeats;
	LV2_URID tracker_position;
	LV2_URID tracker_mode;
	LV2_URID tracker_graph;

	LV2_Log_Log *log;
	LV2_Log_Logger logger;

	nk_pugl_window_t win;

	LV2UI_Controller *controller;
	LV2UI_Write_Function writer;

	PROPS_T(props, MAX_NPROPS);
	plugstate_t state;
	plugstate_t stash;
	uint32_t mode_size;
	uint32_t graph_size;

	float dy;

	char buf [EDIT_MAX];

	struct {
		int row;
		int col;
	} cursor;

	int32_t beats;
	int32_t fractions;
	int32_t n_col;
	int n_row;

	int position;

	int track_modes [MAX_TRACKS];
	track_t tracks [MAX_TRACKS];

	atom_ser_t ser;

	lua_State *L;

	int n_mode;
	char **modes;

	lbase_t lbase;
	lroot_t lroot;
};

extern int luaopen_lpeg(lua_State *L);

static void *
_message_free(message_t *msg)
{
	if(msg)
	{
		free(msg);
	}

	return NULL;
}

static message_t *
_message_alloc(uint32_t size, LV2_URID type, const uint8_t *body)
{
	const uint32_t sz = sizeof(message_t) + size;
	message_t *msg = malloc(sz);
	if(msg)
	{
		msg->atom.size = size;
		msg->atom.type = type;

		if(body)
			memcpy(msg->body, body, size);
		else
			memset(msg->body, 0x0, size);

		return msg;
	}

	return NULL;
}

static void
_messages_alloc(plughandle_t *handle, int32_t beats, int32_t fractions, int32_t n_col)
{
	handle->beats = beats;
	handle->fractions = fractions;
	handle->n_col = n_col;
	handle->n_row = handle->beats * handle->fractions;

	for(int col = 0; col < MAX_TRACKS; col++)
	{
		track_t *track = &handle->tracks[col];

		track->messages = calloc(handle->n_row, sizeof(message_t *));
	}
}

static void
_messages_free(plughandle_t *handle)
{
	for(int col = 0; col < MAX_TRACKS; col++)
	{
		track_t *track = &handle->tracks[col];

		for(int row = 0; row < handle->n_row; row++)
		{
			track->messages[row] = _message_free(track->messages[row]);
		}

		free(track->messages);
		track->messages = NULL;
	}
}

static void
_messages_realloc(plughandle_t *handle, int32_t beats, int32_t fractions, int32_t n_col)
{
	const int32_t n_row = beats * fractions;

	for(int col = 0; col < MAX_TRACKS; col++)
	{
		track_t *track = &handle->tracks[col];
		message_t **messages = calloc(n_row, sizeof(message_t *));

		for(int row = 0; row < handle->n_row; row++)
		{
			message_t *msg = track->messages[row];

			if(msg) // move old message over to new place
			{
				const double beat_time = (double)row / handle->fractions;
				const int row2 = round(beat_time * fractions);

				if( (row2 >= 0) && (row2 < n_row) )
				{
					if(messages[row2])
						messages[row2] = _message_free(messages[row2]);

					messages[row2] = _message_alloc(msg->atom.size, msg->atom.type, msg->body);
				}
			}

			track->messages[row] = _message_free(msg);
		}

		free(track->messages); // free old messages
		track->messages = messages; // swap new with old messages
	}

	handle->beats = beats;
	handle->fractions = fractions;
	handle->n_col = n_col;
	handle->n_row = n_row;
}

static void
_load_state(plughandle_t *handle, uint32_t size, const void *body)
{
	// clear all messages
	for(int col = 0; col < handle->n_col; col++)
	{
		track_t *track = &handle->tracks[col];

		for(int row = 0; row < handle->n_row; row++)
		{
			track->messages[row] = _message_free(track->messages[row]);
		}
	}

	handle->n_col = 0;
	LV2_ATOM_TUPLE_BODY_FOREACH(body, size, item)
	{
		track_t *track = &handle->tracks[handle->n_col++];
		const LV2_Atom_Sequence *seq = (const LV2_Atom_Sequence *)item;

		LV2_ATOM_SEQUENCE_FOREACH(seq, ev)
		{
			const LV2_Atom *atom = &ev->body;
			const int row = round(ev->time.beats * handle->fractions);

			track->messages[row] = _message_alloc(atom->size, atom->type,
				LV2_ATOM_BODY_CONST(atom));
		}
	}
}

static void
_intercept(void *data, int64_t frames, props_impl_t *impl)
{
	plughandle_t *handle = data;

	_messages_realloc(handle, handle->state.beats, handle->state.fractions, handle->n_col);
}

static void
_intercept_position(void *data, int64_t frames, props_impl_t *impl)
{
	plughandle_t *handle = data;

	const int position = handle->state.position / handle->beats * handle->n_row;
	if(position != handle->position)
	{
		handle->position = position;
		nk_pugl_post_redisplay(&handle->win);
	}
}

static void
_intercept_mode(void *data, int64_t frames, props_impl_t *impl)
{
	plughandle_t *handle = data;

	handle->mode_size = impl->value.size;

	int col = 0;
	LV2_ATOM_TUPLE_BODY_FOREACH(handle->state.mode, handle->mode_size, itm)
	{
		const LV2_Atom_Int *track_mode = (const LV2_Atom_Int *)itm;

		handle->track_modes[col++] = track_mode->body;
	}
}

static void
_intercept_graph(void *data, int64_t frames, props_impl_t *impl)
{
	plughandle_t *handle = data;

	handle->graph_size = impl->value.size;
	_load_state(handle, handle->graph_size, handle->state.graph);
}

static const props_def_t defs [MAX_NPROPS] = {
	{
		.property = TRACKER__offset,
		.offset = offsetof(plugstate_t, offset),
		.type = LV2_ATOM__Int
	},
	{
		.property = TRACKER__beats,
		.offset = offsetof(plugstate_t, beats),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept
	},
	{
		.property = TRACKER__fractions,
		.offset = offsetof(plugstate_t, fractions),
		.type = LV2_ATOM__Int,
		.event_cb = _intercept
	},
	{
		.property = TRACKER__repeats,
		.offset = offsetof(plugstate_t, repeats),
		.type = LV2_ATOM__Int
	},
	{
		.property = TRACKER__position,
		.offset = offsetof(plugstate_t, position),
		.type = LV2_ATOM__Double,
		.event_cb = _intercept_position
	},
	{
		.property = TRACKER__mode,
		.offset = offsetof(plugstate_t, mode),
		.type = LV2_ATOM__Tuple,
		.max_size = MODE_SIZE,
		.event_cb = _intercept_mode
	},
	{
		.property = TRACKER__graph,
		.offset = offsetof(plugstate_t, graph),
		.type = LV2_ATOM__Tuple,
		.max_size = GRAPH_SIZE,
		.event_cb = _intercept_graph
	}
};

static void
_modes_add(plughandle_t *handle, const char *key)
{
	handle->n_mode += 1;
	handle->modes = realloc(handle->modes, handle->n_mode * sizeof(char *));
	handle->modes[handle->n_mode-1] = strdup(key);
}

static void
_modes_free(plughandle_t *handle)
{
	for(int m = 0; m < handle->n_mode; m++)
	{
		if(handle->modes[m])
			free(handle->modes[m]); // from strdup
	}
	free(handle->modes);

	handle->n_mode = 0;
	handle->modes = NULL;
}

static void
_modes_write(plughandle_t *handle, struct nk_context *ctx, int track_mode,
	message_t *msg)
{
	lua_State *L = handle->L;

	if(lua_getglobal(L, "_P") == LUA_TTABLE)
	{
		if(lua_rawgeti(L, -1, track_mode + 1) == LUA_TTABLE)
		{
			if(lua_getfield(L, -1, "write") == LUA_TFUNCTION)
			{
				if(msg)
				{
					latom_t *latom = latom_newuserdata(L, NULL);
					latom->atom = &msg->atom;
					latom->body.raw = msg->body;
				}
				else
				{
					lua_pushnil(L);
				}

				if(lua_pcall(L, 1, 3, 0))
				{
					fprintf(stderr, "_modes_write: %s\n", lua_tostring(L, -1));
					lua_pop(L, 1); // err message
				}
				else
				{
					const bool left_aligned = lua_toboolean(L, -1);
					const uint32_t col = lua_isinteger(L, -2)
						? lua_tointeger(L, -2)
						: 0xffffffff;
					const struct nk_color color = nk_rgba(
						(col >> 24) & 0xff,
						(col >> 16) & 0xff,
						(col >>  8) & 0xff,
						(col >>  0) & 0xff);
					const char *result = lua_isstring(L, -3)
						? lua_tostring(L, -3)
						: "err"; //FIXME
					nk_label_colored(ctx, result,
						left_aligned ? NK_TEXT_LEFT : NK_TEXT_CENTERED, color);
					lua_pop(L, 3); // result, color, left_aligned
				}
			}
			else
				lua_pop(L, 1); // nil
		}
		lua_pop(L, 1); // mode
	}
	lua_pop(L, 1); // _P
}

static void
_modes_edit(plughandle_t *handle, struct nk_context *ctx, int track_mode,
	char *buf, size_t capacity, message_t *msg)
{
	lua_State *L = handle->L;

	if(lua_getglobal(L, "_P") == LUA_TTABLE)
	{
		if(lua_rawgeti(L, -1, track_mode + 1) == LUA_TTABLE)
		{
			if(lua_getfield(L, -1, "edit") == LUA_TFUNCTION)
			{
				if(msg)
				{
					latom_t *latom = latom_newuserdata(L, NULL);
					latom->atom = &msg->atom;
					latom->body.raw = msg->body;
				}
				else
					lua_pushnil(L);

				if(lua_pcall(L, 1, 1, 0))
				{
					fprintf(stderr, "_modes_edit: %s\n", lua_tostring(L, -1));
					lua_pop(L, 1); // err message
				}
				else
				{
					const char *result = lua_isstring(L, -1)
						? lua_tostring(L, -1)
						: NULL;
					if(result)
						strncpy(buf, result, capacity);
					lua_pop(L, 1); // result
				}
			}
			else
				lua_pop(L, 1); // nil
		}
		lua_pop(L, 1); // mode
	}
	lua_pop(L, 1); // _P
}

static void
_modes_parse(plughandle_t *handle, struct nk_context *ctx, int track_mode,
	char *buf, message_t **ptr)
{
	lua_State *L = handle->L;

	if(lua_getglobal(L, "_P") == LUA_TTABLE)
	{
		if(lua_rawgeti(L, -1, track_mode + 1) == LUA_TTABLE)
		{
			if(lua_getfield(L, -1, "parse") == LUA_TFUNCTION)
			{
				atom_ser_t ser = {
					.offset = 0,
					.size = 128,
					.buf = calloc(128, sizeof(uint8_t))
				};

				if(ser.buf)
				{
					lv2_atom_forge_set_sink(&handle->lbase.forge, _sink, _deref, &ser);

					lforge_t *lforge = lforge_newuserdata(L);
					lforge->depth = 0;
					lforge->last.frames = 0;
					lforge->forge = &handle->lbase.forge;

					lua_pushstring(L, buf);

					if(lua_pcall(L, 2, 0, 0))
					{
						fprintf(stderr, "_modes_parse: %s\n", lua_tostring(L, -1));
						lua_pop(L, 1); // err message
					}
					else
					{
						if( (ser.atom->size > 0) && (ser.atom->type) )
						{
							message_t *msg = *ptr;

							_message_free(msg);
							msg = _message_alloc(ser.atom->size, ser.atom->type, LV2_ATOM_BODY_CONST(ser.atom));

							*ptr = msg;
						}
					}

					free(ser.buf);
				}
			}
			else
				lua_pop(L, 1); // nil
		}
		lua_pop(L, 1); // mode
	}
	lua_pop(L, 1); // _P
}

static void
_set_property(plughandle_t *handle, LV2_URID property)
{
	atom_ser_t *ser = &handle->ser;
	ser->offset = 0;
	lv2_atom_forge_set_sink(&handle->forge, _sink, _deref, ser);

	LV2_Atom_Forge_Ref ref = 1;
	props_set(&handle->props, &handle->forge, 0, property, &ref);

	const LV2_Atom_Event *ev = ser->ev;
	const LV2_Atom *atom = &ev->body;
	handle->writer(handle->controller, 0, lv2_atom_total_size(atom),
		handle->atom_eventTransfer, atom);
}

static void
_save_state(plughandle_t *handle)
{
	LV2_Atom_Forge *forge = &handle->forge;
	LV2_Atom_Forge_Frame tup_frame;

	atom_ser_t *ser = &handle->ser;

	// serialize graph
	ser->offset = 0;
	lv2_atom_forge_set_sink(forge, _sink, _deref, ser);
	lv2_atom_forge_tuple(forge, &tup_frame);
	{
		for(int col = 0; col < handle->n_col; col++)
		{
			track_t *track = &handle->tracks[col];
			LV2_Atom_Forge_Frame seq_frame;

			lv2_atom_forge_sequence_head(forge, &seq_frame, handle->atom_beatTime);
			{
				for(int row = 0; row < handle->n_row; row++)
				{
					message_t *msg = track->messages[row];

					if(!msg)
						continue; // empty

					double beat_time = (double)row / handle->fractions;
					lv2_atom_forge_beat_time(forge, beat_time);
					lv2_atom_forge_atom(forge, msg->atom.size, msg->atom.type);
					lv2_atom_forge_write(forge, msg->body, msg->atom.size);
				}
			}
			lv2_atom_forge_pop(forge, &seq_frame);
		}
	}
	lv2_atom_forge_pop(forge, &tup_frame);

	const LV2_Atom *atom = ser->atom;
	if(atom->size <= GRAPH_SIZE)
	{
		//FIXME
		props_impl_t *impl = _props_impl_get(&handle->props, handle->tracker_graph);
		if(impl)
		{
			impl->value.size = atom->size;
			memcpy(handle->state.graph, LV2_ATOM_BODY_CONST(atom), atom->size);
		}
	}

	// serialize mode
	ser->offset = 0;
	lv2_atom_forge_set_sink(forge, _sink, _deref, ser);
	lv2_atom_forge_tuple(forge, &tup_frame);
	{
		for(int col = 0; col < handle->n_col; col++)
		{
			const int track_mode = handle->track_modes[col];

			lv2_atom_forge_int(forge, track_mode);
		}
	}
	lv2_atom_forge_pop(forge, &tup_frame);

	atom = ser->atom;
	if(atom->size <= MODE_SIZE)
	{
		//FIXME
		props_impl_t *impl = _props_impl_get(&handle->props, handle->tracker_mode);
		if(impl)
		{
			impl->value.size = atom->size;
			memcpy(handle->state.mode , LV2_ATOM_BODY_CONST(atom), atom->size);
		}
	}

	// update beats and fractions
	handle->state.beats = handle->beats;
	handle->state.fractions= handle->fractions;

	// send all patch:Set messages
	_set_property(handle, handle->tracker_offset);
	_set_property(handle, handle->tracker_beats);
	_set_property(handle, handle->tracker_fractions);
	_set_property(handle, handle->tracker_repeats);
	_set_property(handle, handle->tracker_mode);
	_set_property(handle, handle->tracker_graph);
}

static void
_restore_state(plughandle_t *handle)
{
	atom_ser_t *ser = &handle->ser;
	ser->offset = 0;
	lv2_atom_forge_set_sink(&handle->forge, _sink, _deref, ser);

	LV2_Atom_Forge_Frame frame;
	lv2_atom_forge_object(&handle->forge, &frame, 0, handle->props.urid.patch_get);
	lv2_atom_forge_pop(&handle->forge, &frame);

	const LV2_Atom *atom = ser->atom;
	handle->writer(handle->controller, 0, lv2_atom_total_size(atom),
		handle->atom_eventTransfer, atom);
}

static bool
_tooltip_visible(struct nk_context *ctx)
{
	return nk_widget_has_mouse_click_down(ctx, NK_BUTTON_RIGHT, nk_true)
		|| (nk_widget_is_hovered(ctx) && nk_input_is_key_down(&ctx->input, NK_KEY_CTRL));
}

static inline void
_draw_separator(struct nk_context *ctx, float line_width)
{
	const struct nk_rect b = nk_widget_bounds(ctx);
	const float x0 = b.x;
	const float x1 = b.x + b.w;
	const float y = b.y + b.h;
	struct nk_command_buffer *canvas = nk_window_get_canvas(ctx);
	nk_stroke_line(canvas, x0, y, x1, y, line_width, ctx->style.window.background);
}

static void
_expose(struct nk_context *ctx, struct nk_rect wbounds, void *data)
{
	plughandle_t *handle = data;

	handle->dy = 20.f * nk_pugl_get_scale(&handle->win);
	const float dy = handle->dy;
	const struct nk_vec2 window_padding = ctx->style.window.padding;
	const struct nk_vec2 group_padding = ctx->style.window.group_padding;
	struct nk_style *style = &ctx->style;
	const float header_h = 2*dy + style->window.scrollbar_size.y + 3*group_padding.y;
	const float footer_h = 2*dy + style->window.scrollbar_size.y + 3*group_padding.y;
	const float body_h = wbounds.h - 4*window_padding.y - header_h - footer_h;
	const float strip_h = body_h - style->window.scrollbar_size.y;

	bool needs_realloc = false;
	int32_t n_col_new;
	int32_t n_frac_new;
	int32_t n_beat_new;

	// intercept user key presses
	struct nk_input *in = &ctx->input;
	const bool has_control_q = nk_pugl_is_shortcut_pressed(in, 'q', true);
	const bool has_control_l = nk_pugl_is_shortcut_pressed(in, 'l', true);
	const bool has_control_s = nk_pugl_is_shortcut_pressed(in, 's', true);

	if(has_control_q) // user wants to quit editing mode
		handle->cursor.row = -1;
	if(has_control_l) // user wants to restore state from DSP
		_restore_state(handle);
	if(has_control_s) // user wants to save state to DSP
		_save_state(handle);

	const char *window_name = "Tracker";
	if(nk_begin(ctx, window_name, wbounds, NK_WINDOW_NO_SCROLLBAR))
	{
		struct nk_command_buffer *canvas = nk_window_get_canvas(ctx);

		nk_layout_row_dynamic(ctx, header_h , 1);
		if(nk_group_begin(ctx, "header", 0))
		{
			const float dx = 1.f / (handle->n_col*2 + 2);

			nk_layout_row_begin(ctx, NK_DYNAMIC, dy, handle->n_col + 2);
			{
				nk_layout_row_push(ctx, dx);
				nk_spacing(ctx, 1);

				nk_layout_row_push(ctx, dx);
				nk_spacing(ctx, 1);

				for(int col = 0; col < handle->n_col; col++)
				{
					nk_layout_row_push(ctx, dx*2);
					int track_mode = handle->track_modes[col];
					track_mode = nk_combo(ctx, (const char **)handle->modes,
						handle->n_mode, track_mode,
						dy, nk_vec2(nk_widget_width(ctx), dy*5));
					handle->track_modes[col] = track_mode;
				}
			}
			nk_layout_row_end(ctx);

			nk_layout_row_begin(ctx, NK_DYNAMIC, dy, handle->n_col + 2);
			{
				nk_layout_row_push(ctx, dx);
				struct nk_rect b1 = nk_widget_bounds(ctx);
				nk_fill_rect(canvas, b1, 4.f, nk_rgb(0x10, 0x10, 0x10));
				nk_label(ctx, "Beat", NK_TEXT_CENTERED);

				nk_layout_row_push(ctx, dx);
				struct nk_rect b2 = nk_widget_bounds(ctx);
				nk_fill_rect(canvas, b2, 4.f, nk_rgb(0x10, 0x10, 0x10));
				nk_label(ctx, "Frac", NK_TEXT_CENTERED);

				for(int col = 0; col < handle->n_col; col++)
				{
					nk_layout_row_push(ctx, dx*2);
					struct nk_rect b3 = nk_widget_bounds(ctx);
					nk_fill_rect(canvas, b3, 4.f, nk_rgb(0x20, 0x20, 0x20));
					nk_labelf(ctx, NK_TEXT_CENTERED, "Track %"PRIi32, (int32_t)col + 1);
				}
			}
			nk_layout_row_end(ctx);

			nk_group_end(ctx);
		}

		nk_layout_row_dynamic(ctx, body_h, 1);
		struct nk_list_view lview;
		if(nk_list_view_begin(ctx, &lview, "body", 0, dy, handle->n_row))
		{
			for(int row = lview.begin; (row < lview.end) && (row < handle->n_row); row++)
			{
				const bool draw_strip = row == lview.begin;
				const bool draw_separator = ( (row + 1) % handle->fractions) == 0;

				const int32_t beats = round((double)row / handle->fractions);
				const int32_t position = row % handle->fractions;

				const float dx = 1.f / (handle->n_col*2 + 2);
				nk_layout_row_begin(ctx, NK_DYNAMIC, dy, handle->n_col + 2);

				{
					nk_layout_row_push(ctx, dx);
					if(draw_strip)
					{
						struct nk_rect b1 = nk_widget_bounds(ctx);
						b1.y += nk_window_get_panel(ctx)->offset_y[0] - group_padding.y;
						b1.h = strip_h;
						nk_fill_rect(canvas, b1, 4.f, nk_rgb(0x10, 0x10, 0x10));
					}
					if(draw_separator)
					{
						_draw_separator(ctx, 3.f);
					}
					if(row == handle->position)
					{
						nk_fill_rect(canvas, nk_widget_bounds(ctx), 4.f, nk_rgb(0x40, 0x40, 0x40));
					}

					if(position == 0)
						nk_labelf(ctx, NK_TEXT_CENTERED, "%"PRIi32, handle->state.offset + beats);
					else
						nk_spacing(ctx, 1);

					nk_layout_row_push(ctx, dx);
					if(draw_strip)
					{
						struct nk_rect b1 = nk_widget_bounds(ctx);
						b1.y += nk_window_get_panel(ctx)->offset_y[0] - group_padding.y;
						b1.h = strip_h;
						nk_fill_rect(canvas, b1, 4.f, nk_rgb(0x10, 0x10, 0x10));
					}
					_draw_separator(ctx, draw_separator ? 3.f : 1.f);
					if(row == handle->position)
					{
						nk_fill_rect(canvas, nk_widget_bounds(ctx), 4.f, nk_rgb(0x40, 0x40, 0x40));
					}

					nk_labelf(ctx, NK_TEXT_CENTERED, "%"PRIi32, position + 1);
				}

				for(int col = 0; col < handle->n_col; col++)
				{
					message_t **msg = &handle->tracks[col].messages[row];
					const int track_mode = handle->track_modes[col];

					nk_layout_row_push(ctx, dx*2);
					if(draw_strip)
					{
						struct nk_rect b1 = nk_widget_bounds(ctx);
						b1.y += nk_window_get_panel(ctx)->offset_y[0] - group_padding.y;
						b1.h = strip_h;
						nk_fill_rect(canvas, b1, 4.f, nk_rgb(0x20, 0x20, 0x20));
					}
					_draw_separator(ctx, draw_separator ? 3.f : 1.f);
					if(row == handle->position)
					{
						nk_fill_rect(canvas, nk_widget_bounds(ctx), 4.f, nk_rgb(0x40, 0x40, 0x40));
					}

					// handle user entering edit mode
					if(nk_widget_is_mouse_clicked(ctx, NK_BUTTON_LEFT))
					{
						_modes_edit(handle, ctx, track_mode, handle->buf, EDIT_MAX, *msg);

						handle->cursor.row = row;
						handle->cursor.col = col;

						nk_pugl_post_redisplay(&handle->win);
					}
					else if(nk_widget_is_mouse_clicked(ctx, NK_BUTTON_RIGHT))
					{
						*msg = _message_free(*msg);
					}

					if( (row == handle->cursor.row) && (col == handle->cursor.col) )
					{
						nk_edit_focus(ctx, NK_EDIT_SIMPLE);
						ctx->current->edit.mode = NK_TEXT_EDIT_MODE_REPLACE;

						const nk_flags flags = nk_edit_string_zero_terminated(ctx,
							NK_EDIT_SIMPLE, handle->buf, EDIT_MAX,
							nk_filter_default);
						(void)flags;

						//FIXME track_driver->action(handle, ctx);
					}
					else
					{
						_modes_write(handle, ctx, track_mode, *msg);
					}
				}

				nk_layout_row_end(ctx);
			}

			nk_list_view_end(&lview);
		}

		nk_layout_row_dynamic(ctx, footer_h, 1);
		if(nk_group_begin(ctx, "footer", 0))
		{
			nk_layout_row_dynamic(ctx, dy, 5);

			const int n_col1 = handle->n_col;
			n_col_new = nk_propertyi(ctx, "Tracks:", 1, n_col1, MAX_TRACKS, 1, 1.f);
			if(n_col1 != n_col_new)
			{
				needs_realloc = true;
			}

			handle->state.offset = nk_propertyi(ctx, "Offset:",
				1, handle->state.offset, 1024 - handle->beats, 1, 1.f);

			const int n_beat1 = handle->beats;
			n_beat_new = nk_propertyi(ctx, "Beats:", 1, n_beat1, 1024, 1, 1.f);
			if(n_beat1 != n_beat_new)
			{
				needs_realloc = true;
			}

			const int n_frac1 = handle->fractions;
			n_frac_new = nk_propertyi(ctx, "Fractions:", 1, n_frac1, 64, 1, 1.f);
			if(n_frac1 != n_frac_new)
			{
				needs_realloc = true;
			}

			handle->state.repeats = nk_propertyi(ctx, "Repeats:",
				0, handle->state.repeats, 1024, 1, 1.f);

			nk_layout_row_dynamic(ctx, dy, 3);

			//const bool load_commited = true; //FIXME
			if(_tooltip_visible(ctx))
				nk_tooltip(ctx, "Ctrl-L");
			/*FIXME
			nk_style_push_style_item(ctx, &ctx->style.button.normal, load_commited
				? nk_style_item_color(nk_default_color_style[NK_COLOR_BUTTON_ACTIVE])
				: nk_style_item_color(nk_default_color_style[NK_COLOR_BUTTON]));
			*/
			if(nk_button_label(ctx, "Load"))
				_restore_state(handle);
			/*FIXME
			nk_style_pop_style_item(ctx);
			*/

			//const bool save_commited = true; //FIXME
			if(_tooltip_visible(ctx))
				nk_tooltip(ctx, "Ctrl-S");
			/*FIXME
			nk_style_push_style_item(ctx, &ctx->style.button.normal, save_commited
				? nk_style_item_color(nk_default_color_style[NK_COLOR_BUTTON_ACTIVE])
				: nk_style_item_color(nk_default_color_style[NK_COLOR_BUTTON]));
			*/
			if(nk_button_label(ctx, "Save"))
				_save_state(handle);
			/*FIXME
			nk_style_pop_style_item(ctx);
			*/

			nk_labelf(ctx, NK_TEXT_RIGHT, "Tracker.lv2: "TRACKER_VERSION);

			nk_group_end(ctx);
		}
	}

	// handle user navigating in edit mode
	const bool has_enter = nk_input_is_key_pressed(in, NK_KEY_ENTER);
	const bool has_tab = nk_input_is_key_pressed(in, NK_KEY_TAB);

	if( (handle->cursor.row != -1) && (has_enter || has_tab) )
	{
		// parse midi from user input
		{
			message_t **msg = &handle->tracks[handle->cursor.col].messages[handle->cursor.row];
			const int track_mode = handle->track_modes[handle->cursor.col];

			_modes_parse(handle, ctx, track_mode, handle->buf, msg);
			ctx->current->edit.cursor = 0; // reset pointer
		}

		// advance cursor right
		if(has_tab)
		{
			handle->cursor.col += 1;

			if(handle->cursor.col >= handle->n_col) // overflowed
			{
				handle->cursor.row += 1;
				handle->cursor.col = 0;

				if(handle->cursor.row >= handle->n_row) // overflowed
				{
					handle->cursor.row = -1;
				}
			}
		}

		// advance cursor down
		if(has_enter)
		{
			handle->cursor.row += 1;

			if(handle->cursor.row >= handle->n_row) // overflowed
			{
				handle->cursor.row = -1;
			}
		}

		// prepare new cursor
		if(handle->cursor.row != -1)
		{
			message_t **msg = &handle->tracks[handle->cursor.col].messages[handle->cursor.row];
			const int track_mode = handle->track_modes[handle->cursor.col];

			_modes_edit(handle, ctx, track_mode, handle->buf, EDIT_MAX, *msg);
		}
	}

	nk_end(ctx);

	if(needs_realloc)
	{
		_messages_realloc(handle, n_beat_new, n_frac_new, n_col_new);
		nk_pugl_post_redisplay(&handle->win);
	}
}

static void
_reload(plughandle_t *handle)
{
	lua_State *L = handle->L;

	_modes_free(handle);

	if(lua_getglobal(L, "_P") == LUA_TTABLE)
	{
		lua_pushnil(L);
		while(lua_next(L, -2)) // -2: key, -1: value
		{
			if(lua_getfield(L, -1, "label") == LUA_TSTRING)
			{
				const char *label = lua_tostring(L, -1);
				_modes_add(handle, label);
			}
			lua_pop(L, 2); // label, value
		}
	}
	lua_pop(L, 1);
}

static int
_lmap(lua_State *L)
{
	plughandle_t *handle = lua_touserdata(L, lua_upvalueindex(1));
	const char *uri = luaL_checkstring(L, 1);

	lua_pushinteger(L, handle->map->map(handle->map->handle, uri));

	return 1;
}

static int
_lunmap(lua_State *L)
{
	plughandle_t *handle = lua_touserdata(L, lua_upvalueindex(1));
	const LV2_URID urid = luaL_checkinteger(L, 1);

	lua_pushstring(L, handle->unmap->unmap(handle->unmap->handle, urid));

	return 1;
}

static const luaL_Reg globals [] = {
	{"Map", _lmap},
	{"Unmap", _lunmap},

	{NULL, NULL}
};

static LV2UI_Handle
instantiate(const LV2UI_Descriptor *descriptor, const char *plugin_uri,
	const char *bundle_path, LV2UI_Write_Function write_function,
	LV2UI_Controller controller, LV2UI_Widget *widget,
	const LV2_Feature *const *features)
{
	plughandle_t *handle = calloc(1, sizeof(plughandle_t));
	if(!handle)
		return NULL;

	void *parent = NULL;
	LV2UI_Resize *host_resize = NULL;
	for(int i=0; features[i]; i++)
	{
		if(!strcmp(features[i]->URI, LV2_UI__parent))
			parent = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_UI__resize))
			host_resize = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_URID__map))
			handle->map = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_URID__unmap))
			handle->unmap = features[i]->data;
		else if(!strcmp(features[i]->URI, LV2_LOG__log))
			handle->log = features[i]->data;
	}

	if(!parent)
	{
		fprintf(stderr,
			"%s: Host does not support ui:parent\n", descriptor->URI);
		free(handle);
		return NULL;
	}
	if(!handle->map)
	{
		fprintf(stderr,
			"%s: Host does not support urid:map\n", descriptor->URI);
		free(handle);
		return NULL;
	}

	if(handle->log)
		lv2_log_logger_init(&handle->logger, handle->map, handle->log);

	lv2_atom_forge_init(&handle->forge, handle->map);

	lbase_init(&handle->lbase, plugin_uri, handle->map);
	lroot_init(&handle->lroot, handle->map);

	if(!props_init(&handle->props, plugin_uri,
		defs, MAX_NPROPS, &handle->state, &handle->stash,
		handle->map, handle))
	{
		fprintf(stderr, "props_init failed\n");
		free(handle);
		return NULL;
	}

	handle->atom_eventTransfer = handle->map->map(handle->map->handle, LV2_ATOM__eventTransfer);
	handle->atom_beatTime = handle->map->map(handle->map->handle, LV2_ATOM__beatTime);
	handle->midi_MidiEvent = handle->map->map(handle->map->handle, LV2_MIDI__MidiEvent);
	handle->tracker_offset = handle->map->map(handle->map->handle, TRACKER__offset);
	handle->tracker_beats = handle->map->map(handle->map->handle, TRACKER__beats);
	handle->tracker_fractions = handle->map->map(handle->map->handle, TRACKER__fractions);
	handle->tracker_repeats = handle->map->map(handle->map->handle, TRACKER__repeats);
	handle->tracker_position = handle->map->map(handle->map->handle, TRACKER__position);
	handle->tracker_mode = handle->map->map(handle->map->handle, TRACKER__mode);
	handle->tracker_graph = handle->map->map(handle->map->handle, TRACKER__graph);

	handle->controller = controller;
	handle->writer = write_function;

	nk_pugl_config_t *cfg = &handle->win.cfg;
	cfg->width = 720;
	cfg->height = 720;
	cfg->resizable = true;
	cfg->ignore = false;
	cfg->class = "tracker";
	cfg->title = "Tracker";
	cfg->parent = (intptr_t)parent;
	cfg->host_resize = host_resize;
	cfg->data = handle;
	cfg->expose = _expose;

	if(asprintf(&cfg->font.face, "%sCousine-Regular.ttf", bundle_path) == -1)
		cfg->font.face = NULL;
	cfg->font.size = 13;
	
	*(intptr_t *)widget = nk_pugl_init(&handle->win);
	nk_pugl_show(&handle->win);

	handle->L = luaL_newstate(); //FIXME check
	if(handle->L)
	{
		luaL_requiref(handle->L, "base", luaopen_base, 0);
		luaL_requiref(handle->L, "coroutine", luaopen_coroutine, 1);
		luaL_requiref(handle->L, "table", luaopen_table, 1);
		luaL_requiref(handle->L, "string", luaopen_string, 1);
		luaL_requiref(handle->L, "math", luaopen_math, 1);
		luaL_requiref(handle->L, "utf8", luaopen_utf8, 1);
		luaL_requiref(handle->L, "debug", luaopen_debug, 1);

		luaL_requiref(handle->L, "lpeg", luaopen_lpeg, 1);
		lua_pop(handle->L, 8);

		//luaL_requiref(handle->L, "io", luaopen_io, 1);
		//luaL_requiref(handle->L, "os", luaopen_os, 1);
		//luaL_requiref(handle->L, "bit32", luaopen_bit32, 1);
		//luaL_requiref(handle->L, "package", luaopen_package, 1);

		// automatic garbage collector
		lua_gc(handle->L, LUA_GCRESTART, 0); // enable automatic garbage collection
		lua_gc(handle->L, LUA_GCSETPAUSE, 105); // next step when memory increased by 5%
		lua_gc(handle->L, LUA_GCSETSTEPMUL, 105); // run 5% faster than memory allocation

		// register forge metatables
		luaL_newmetatable(handle->L, "lforge");
		lua_pushlightuserdata(handle->L, &handle->lbase); // @ upvalueindex 1
		luaL_setfuncs(handle->L, lforge_mt, 1);
		// protect metatable
		lua_pushboolean(handle->L, 0); \
		lua_setfield(handle->L, -2, "__metatable");
		// index metatable
		lua_pushvalue(handle->L, -1); \
		lua_setfield(handle->L, -2, "__index");
		lua_pop(handle->L, 1);

		// register atom metatables
		luaL_newmetatable(handle->L, "latom");
		lua_pushlightuserdata(handle->L, &handle->lroot); // @ upvalueindex 1
		luaL_setfuncs(handle->L, latom_mt, 1);
		// protect metatable
		lua_pushboolean(handle->L, 0); \
		lua_setfield(handle->L, -2, "__metatable");

		lua_pushglobaltable(handle->L);
		lua_pushlightuserdata(handle->L, handle);
		luaL_setfuncs(handle->L, globals, 1);

		char *path;
		if(asprintf(&path, "%stracker.lua", bundle_path) == -1)
			path = NULL;
		if(path)
		{
			if(luaL_dofile(handle->L, path))
				fprintf(stderr, "Lua: %s\n", lua_tostring(handle->L, -1));
			free(path);

			_reload(handle);
		};
		for(int t = 0; t < MAX_TRACKS; t++)
		{
			handle->tracks[t].L = lua_newthread(handle->L);
			handle->tracks[t].ref = luaL_ref(handle->L, LUA_REGISTRYINDEX);
		}
	}

	handle->cursor.row = -1;
	handle->cursor.col = -1;

	//FIXME
	handle->state.offset = 1;
	handle->state.repeats = 0;
	_messages_alloc(handle, 4, 8, 4);

	atom_ser_t *ser = &handle->ser;
	ser->size = 1024;
	ser->buf = malloc(ser->size);

	_restore_state(handle);

	return handle;
}

static void
cleanup(LV2UI_Handle instance)
{
	plughandle_t *handle = instance;

	if(handle->win.cfg.font.face)
		free(handle->win.cfg.font.face);
	nk_pugl_hide(&handle->win);
	nk_pugl_shutdown(&handle->win);

	_modes_free(handle);
	_messages_free(handle);

	if(handle->L)
	{
		for(int t = 0; t < MAX_TRACKS; t++)
		{
			luaL_unref(handle->L, LUA_REGISTRYINDEX, handle->tracks[t].ref);
		}

		lua_close(handle->L);
	}

	atom_ser_t *ser = &handle->ser;
	if(ser->buf)
		free(ser->buf);

	free(handle);
}

static void
port_event(LV2UI_Handle instance, uint32_t index, uint32_t size,
	uint32_t protocol, const void *buf)
{
	plughandle_t *handle = instance;

	if(protocol == handle->atom_eventTransfer)
	{
		const LV2_Atom_Object *obj = buf;

		atom_ser_t *ser = &handle->ser;
		ser->offset = 0;
		lv2_atom_forge_set_sink(&handle->forge, _sink, _deref, ser);

		LV2_Atom_Forge_Ref ref = 0;
		if(props_advance(&handle->props, &handle->forge, 0, obj, &ref))
		{
			nk_pugl_post_redisplay(&handle->win);
		}
	}
}

static int
_idle(LV2UI_Handle instance)
{
	plughandle_t *handle = instance;

	return nk_pugl_process_events(&handle->win);
}

static const LV2UI_Idle_Interface idle_ext = {
	.idle = _idle
};

static int
_resize(LV2UI_Handle instance, int width, int height)
{
	plughandle_t *handle = instance;

	return nk_pugl_resize(&handle->win, width, height);
}

static const LV2UI_Resize resize_ext = {
	.ui_resize = _resize
};

static const void *
extension_data(const char *uri)
{
	if(!strcmp(uri, LV2_UI__idleInterface))
		return &idle_ext;
	else if(!strcmp(uri, LV2_UI__resize))
		return &resize_ext;
		
	return NULL;
}

const LV2UI_Descriptor tracker_tracker_ui = {
	.URI            = TRACKER_PREFIX"tracker_ui",
	.instantiate    = instantiate,
	.cleanup        = cleanup,
	.port_event     = port_event,
	.extension_data = extension_data
};
